
/******************************************************************************
* MODULE     : parsehtml.cc
* DESCRIPTION: conversion of xml and html strings into logical html trees
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "convert.hh"
#include "hashset.hh"

/******************************************************************************
* The xml/html parser aims to parse a superset of the set of valid documents.
* In other words, no attempts are made to signal error messages for
* incorrect documents; in the case of Html we even attempt to correct
* common mistakes, like badly structured documents. So correct documents
* should be parsed correctly and incorrect documents are transformed into
* correct documents in a heuristic way.
*
* The parser proceeds in three steps: the first pass does all parsing
* except for the construction of a tree structure for nested tags.
* The second stage takes care of the nesting, while heuristically
* correcting improper nested trees, and while taking care of optional
* closing tags in the case of Html. The last stage does some final
* white space and entity cleanup.
*
* Present limitations: we do not really parse <!DOCTYPE ...> constructs yet
* Consequently, we perform no entity replacements and it is not yet possible
* to associate default xml:space attributes to tags.
******************************************************************************/

struct xml_html_parser {
  bool html;
  int i, n;
  string s;
  array<tree> a;
  tree stack;

  xml_html_parser ();
  inline bool is_space (char c) {
    return (c==' ') || (c=='\11') || (c=='\12') || (c=='\15'); }
  inline void skip_space () {
    while ((i<n) && is_space (s[i])) i++; }
  inline bool is_name_char (char c) {
    return is_alpha (c) || is_digit (c) ||
      (c == '_') || (c == ':') || (c == '.') || (c == '-') ||
      (((int) ((unsigned char) c)) >= 128); }

  string parse_until (string what);
  string parse_name ();
  string parse_entity ();
  tree parse_attribute ();
  tree parse_opening ();
  tree parse_closing ();
  tree parse_pi ();
  tree parse_comment ();
  tree parse_cdata ();
  tree parse_doctype ();
  tree parse_misc ();
  void parse ();

  // NOTE: these routines should remain there even if they are not used
  bool finalize_preserve_space (string tag);
  string finalize_space (string s, bool first, bool last);
  tree finalize_space (tree t);
  // END NOTE
  bool build_valid_child (string parent, string child);
  bool build_must_close (string tag);
  bool build_can_close (string tag);
  void build (tree& r);

  tree finalize_entities (tree t);
  tree finalize_sxml (tree t);
  tree parse (string s);
};

/******************************************************************************
* Initialization
******************************************************************************/

static hashset<string> html_empty_tag_table;
static hashset<string> html_auto_close_table;
static hashset<string> html_block_table;
static hashmap<string,string> html_entity ("");

xml_html_parser::xml_html_parser () {
  if (N(html_empty_tag_table) == 0) {
    html_empty_tag_table->insert ("basefont");
    html_empty_tag_table->insert ("br");
    html_empty_tag_table->insert ("area");
    html_empty_tag_table->insert ("link");
    html_empty_tag_table->insert ("param");
    html_empty_tag_table->insert ("hr");
    html_empty_tag_table->insert ("input");
    html_empty_tag_table->insert ("col");
    html_empty_tag_table->insert ("frame");
    html_empty_tag_table->insert ("isindex");
    html_empty_tag_table->insert ("base");
    html_empty_tag_table->insert ("meta");
  }

  if (N(html_auto_close_table) == 0) {
    html_auto_close_table->insert ("body");
    html_auto_close_table->insert ("p");
    html_auto_close_table->insert ("dt");
    html_auto_close_table->insert ("dd");
    html_auto_close_table->insert ("li");
    html_auto_close_table->insert ("option");
    html_auto_close_table->insert ("thead");
    html_auto_close_table->insert ("tfoot");
    html_auto_close_table->insert ("tbody");
    html_auto_close_table->insert ("colgroup");
    html_auto_close_table->insert ("tr");
    html_auto_close_table->insert ("th");
    html_auto_close_table->insert ("td");
    html_auto_close_table->insert ("head");
    html_auto_close_table->insert ("html");
  }

  if (N(html_block_table) == 0) {
    html_block_table->insert ("h1");
    html_block_table->insert ("h2");
    html_block_table->insert ("h3");
    html_block_table->insert ("h4");
    html_block_table->insert ("h5");
    html_block_table->insert ("h6");
    html_block_table->insert ("ul");
    html_block_table->insert ("ol");
    html_block_table->insert ("li");
    html_block_table->insert ("dl");
    html_block_table->insert ("dd");
    html_block_table->insert ("dt");
    html_block_table->insert ("pre");
    html_block_table->insert ("div");
    html_block_table->insert ("p");
    html_block_table->insert ("noscript");
    html_block_table->insert ("blockquote");
    html_block_table->insert ("form");
    html_block_table->insert ("hr");
    html_block_table->insert ("table");
    html_block_table->insert ("fieldset");
    html_block_table->insert ("address");
  }

  if (N (html_entity) == 0) {
    html_entity ("quot")= "\"";
    html_entity ("amp")= "&";
    html_entity ("lt")= "<";
    html_entity ("gt")= ">";
    html_entity ("nbsp")= "";

    html_entity ("Agrave")= "";
    html_entity ("Aacute")= "";
    html_entity ("Acirc")= "";
    html_entity ("Atilde")= "";
    html_entity ("Auml")= "";
    html_entity ("Aring")= "";
    html_entity ("AElig")= "";
    html_entity ("Ccedil")= "";
    html_entity ("Egrave")= "";
    html_entity ("Eacute")= "";
    html_entity ("Ecirc")= "";
    html_entity ("Euml")= "";
    html_entity ("Igrave")= "";
    html_entity ("Iacute")= "";
    html_entity ("Icirc")= "";
    html_entity ("Iuml")= "";
    html_entity ("Ntilde")= "";
    html_entity ("Oslash")= "";
    html_entity ("Ograve")= "";
    html_entity ("Oacute")= "";
    html_entity ("Ocirc")= "";
    html_entity ("Otilde")= "";
    html_entity ("Ouml")= "";
    html_entity ("Ugrave")= "";
    html_entity ("Uacute")= "";
    html_entity ("Ucirc")= "";
    html_entity ("Uuml")= "";
    html_entity ("Yacute")= "";
    html_entity ("szlig")= "";

    html_entity ("agrave")= "";
    html_entity ("aacute")= "";
    html_entity ("acirc")= "";
    html_entity ("atilde")= "";
    html_entity ("auml")= "";
    html_entity ("aring")= "";
    html_entity ("aelig")= "";
    html_entity ("ccedil")= "";
    html_entity ("egrave")= "";
    html_entity ("eacute")= "";
    html_entity ("ecirc")= "";
    html_entity ("euml")= "";
    html_entity ("igrave")= "";
    html_entity ("iacute")= "";
    html_entity ("icirc")= "";
    html_entity ("iuml")= "";
    html_entity ("ntilde")= "";
    html_entity ("oslash")= "";
    html_entity ("ograve")= "";
    html_entity ("oacute")= "";
    html_entity ("ocirc")= "";
    html_entity ("otilde")= "";
    html_entity ("ouml")= "";
    html_entity ("ugrave")= "";
    html_entity ("uacute")= "";
    html_entity ("ucirc")= "";
    html_entity ("uuml")= "";
    html_entity ("yacute")= "";
    html_entity ("yuml")= "";
  }
}

/******************************************************************************
* Parsing without structuring
******************************************************************************/

string
xml_html_parser::parse_until (string what) {
  int start= i;
  while ((i<n) && (!test (s, i, what))) i++;
  if (test (s, i, what)) {
    string r= s (start, i);
    i += N(what);
    return r;
  }
  return "";
}

string
xml_html_parser::parse_name () {
  int start= i;
  while ((i<n) && is_name_char (s[i])) i++;
  if (html) return locase_all (s (start, i));
  return s (start, i);
}

string
xml_html_parser::parse_entity () {
  int start= i++;
  if (s[i] == '#') {
    i++;
    while ((i<n) && is_digit (s[i])) i++;
  }
  else (void) parse_name ();
  if ((i<n) && (s[i] == ';')) i++;
  return s (start, i);
}

tree
xml_html_parser::parse_attribute () {
  string attr= parse_name (), val;
  bool no_val= false;
  skip_space ();
  if (test (s, i, "=")) i++;
  skip_space ();
  if (test (s, i, "\42")) {
    i++;
    val= parse_until ("\42");
  }
  else if (test (s, i, "'")) {
    i++;
    val= parse_until ("'");
  }
  else { // for Html
    int start= i;
    while (i<n) {
      if (is_space (s[i]) || (s[i]=='<') || (s[i]=='>')) break;
      i++;
    }
    val= s (start, i);
    no_val= (start == i);
  }
  if (!no_val) return tuple ("attr", attr, val);
  else if (attr != "") return tuple ("attr", attr);
  else return tuple ("attr");
}

tree
xml_html_parser::parse_opening () {
  i++;
  string name= parse_name ();
  tree t= tuple ("begin", name);
  while (true) {
    skip_space ();
    if ((i==n) || (s[i]=='>') || test (s, i, "/>")) break;
    tree attr= parse_attribute ();
    if (attr == tuple ("attr")) break;
    t << attr;
  }
  if (test (s, i, "/>")) { t[0]= "tag"; i+=2; }
  else if (test (s, i, ">")) i++;
  return t;
}

tree
xml_html_parser::parse_closing () {
  i+=2;
  string name= parse_name ();
  (void) parse_until (">");
  return tuple ("end", name);
}

tree
xml_html_parser::parse_pi () {
  i+=2;
  string name= parse_name ();
  skip_space ();
  return tuple ("pi", name, parse_until ("?>"));
}

tree
xml_html_parser::parse_comment () {
  i+=4;
  return tuple ("comment", parse_until ("-->"));
}

tree
xml_html_parser::parse_cdata () {
  i+=9;
  return tuple ("cdata", parse_until ("]]>"));
}

tree
xml_html_parser::parse_doctype () {
  i+=9;
  // FIXME: doctype may contain '>' in ExternalID and in internal DTD subset
  skip_space ();
  return tuple ("doctype", parse_until (">"));
}

tree
xml_html_parser::parse_misc () {
  i+=2;
  tree t= tuple ("misc");
  while (true) {
    skip_space ();
    if (test (s, i, ">")) { i++; break; }
    int start= i;
    while (i<n) {
      char c= s[i];
      if (is_space (c) || (c == '>')) break;
      i++;
    }
    t << s (start, i);
  }
  return t;
}

void
xml_html_parser::parse () {
  int start= i;
  while (i<n) {
    // cout << "Parsing " << i << "/" << n << "\n";
    if (s[i] == '<') {
      if (i>start) a << tree (s (start, i));
      if (test (s, i, "</")) a << parse_closing ();
      else if (test (s, i, "<?")) a << parse_pi ();
      else if (test (s, i, "<!--")) a << parse_comment ();
      else if (test (s, i, "<![CDATA[")) a << parse_cdata ();
      else if (test (s, i, "<!DOCTYPE")) a << parse_doctype ();
      else if (test (s, i, "<!")) a << parse_misc ();
      else a << parse_opening ();
      start= i;
    }
    else if (s[i] == '&')
      (void) parse_entity ();
    else i++;
  }
  if (i>start) a << tree (s (start, i));
}

/******************************************************************************
* Building the structured parse tree with error correction
******************************************************************************/

bool
xml_html_parser::build_valid_child (string parent, string child) {
  if (!html) return true;
  if ((parent == "<bottom>") || (parent == "html") || (parent == "body"))
    return true;
  if (html_empty_tag_table->contains (parent)) return false;
  if (!html_auto_close_table->contains (child)) return true;
  if (parent == "p") return !html_block_table->contains (child);
  if ((child == "dt") || (child == "dd")) return parent == "dl";
  if (child == "li")
    return (parent == "ul") || (parent == "ol") ||
           (parent == "dir") || (parent == "menu");
  if (child == "option") return (parent == "select") || (parent == "optgroup");
  if ((child == "thead") || (child == "tfoot") || (child == "tbody"))
    return parent == "table";
  if (child == "colgroup") return parent == "table";
  if (child == "col") return (parent == "table") || (parent == "colgroup");
  if (child == "tr")
    return (parent == "table") || (parent == "thead") ||
           (parent == "tfoot") || (parent == "tbody");
  if ((child == "th") || (child == "td"))
    return (parent == "tr") ||
           (parent == "table") || (parent == "thead") ||
           (parent == "tfoot") || (parent == "tbody");
  return true;
}

bool
xml_html_parser::build_must_close (string tag) {
  if (build_valid_child (stack[0]->label, tag)) return false;
  // if !html, we have already returned false
  tree counter= stack;
  while (counter != tuple ("<bottom>")) {
    if (build_valid_child (counter[0]->label, tag)) return true;
    counter= counter[1];
  }
  // since <html> and <body> can have any child we only get here when parsing
  // something where both are omitted and we can close nodes up to the root.
  return true;
}

bool
xml_html_parser::build_can_close (string tag) {
  tree counter= stack[1];
  while (counter != tuple ("<bottom>")) {
    if (counter[0]->label == tag) return true;
    counter= counter[1];
  }
  return false;
}

void
xml_html_parser::build (tree& r) {
  while (i<n) {
    if (is_tuple (a[i], "begin")) {
      string name= a[i][1]->label;
      if (build_must_close (name)) return;
      tree sub= copy (a[i]); sub[0]= "tag";
      i++;
      if (html_empty_tag_table->contains (name))
	r << sub;
      else {
	stack= tuple (name, stack);
	build (sub);
	r << sub;
	stack= stack[1];
      }
    }
    else if (is_tuple (a[i], "end")) {
      if (stack[0]->label == a[i][1]->label) { i++; return; }
      if (build_can_close (a[i][1]->label)) return;
      i++;
    }
    else r << a[i++];
  }
}

/******************************************************************************
* Finalization
******************************************************************************/

bool
xml_html_parser::finalize_preserve_space (string tag) {
  return tag == "pre";
}

string
xml_html_parser::finalize_space (string s, bool first, bool last) {
  int i, n= N(s);
  string r;
  bool flag= first;
  for (i=0; i<n; i++)
    if (is_space (s[i])) {
      if (!flag) r << ' ';
      flag= true;
    }
    else {
      r << s[i];
      flag= false;
    }
  n= N(r);
  if (last && (n>0) && (r[n-1] == ' '))
    r->resize (n-1);
  return r;
}

tree
xml_html_parser::finalize_space (tree t) {
  if (is_atomic (t) || (!is_tuple (t, "tag"))) return t;
  else {
    int i, n= N(t);
    tree r= tuple (t[0], t[1]);
    int first= -1, last= -1;
    for (i=2; i<n; i++)
      if (!is_tuple (t[i], "attr")) {
	first= i; break;
      }
    if (!is_tuple (t[n-1], "attr"))
      last= n-1;
    for (i=2; i<n; i++) {
      if (is_atomic (t[i])) {
	if (finalize_preserve_space (t[1]->label)) r << t[i];
	else {
	  string s= finalize_space (t[i]->label, i==2, i==(n-1));
	  if (s != "") r << s;
	}
      }
      else if (is_tuple (t[i], "tag")) r << finalize_space (t[i]);
      else r << t[i];
    }
    return r;
  }
}

tree
xml_html_parser::finalize_entities (tree t) {
  if (is_atomic (t)) {
    string s= t->label;

    if (html) {
      string r;
      int i, n= N(s);
      for (i=0; i<n; i++)
	if (s[i] == '&') {
	  int start= ++i;
	  if (s[i] == '#') {
	    i++;
	    while ((i<n) && is_digit (s[i])) i++;
	  }
	  else while ((i<n) && is_name_char (s[i])) i++;
	  string ss= s (start, i);
	  if ((i>=n) || (s[i]!=';')) i--;
	  if ((ss != "") && (ss[0] == '#')) {
	    char c= (char) as_int (ss (1, N(ss)));
	    r << string (c);
	  }
	  else if (html_entity->contains (ss))
	    r << copy (html_entity [ss]);
	  else r << ("&" * s (start, i+1));
	}
	else r << s[i];
      s= r;
    }
    
    return s;
  }
  else {
    int i, n= N(t);
    tree r (t, n);
    for (i=0; i<n; i++)
      r[i]= finalize_entities (t[i]);
    return r;
  }
}

tree
xml_html_parser::finalize_sxml (tree t) {
  if (!is_tuple (t, "tag")) return ""; // sanity
  int i, n= N(t);
  tree tag = tuple (t[1]);
  if (t[1] == "<document>") tag= tuple ("*TOP*");
  tree attrs = tuple ("@");
  tree content = tuple ();
  for (i=2; i<n; i++)
    if (is_tuple (t[i], "attr")) {
      tree attr;
      if (N(t[i]) == 2) attr= tuple (t[i][1]);
      else attr= tuple (t[i][1]->label, quote (t[i][2]->label));
      attrs << attr;
    }
    else if (is_tuple (t[i], "tag"))
      content << finalize_sxml (t[i]);
    else if (is_atomic (t[i]))
      content << quote (t[i]->label);
    else if (is_tuple (t[i], "pi"))
      content << tuple ("*PI*", t[i][1]->label, quote (t[i][2]->label));
    else if (is_tuple (t[i], "doctype"))
      content << tuple ("*DOCTYPE*", quote (t[i][1]->label));
    else if (is_tuple (t[i], "cdata"))
      content << quote (t[i][1]->label);
  if (N(attrs) > 1) tag << attrs;
  tag << A(content);
  return tag;
}

/******************************************************************************
* Building the structured parse tree with error correction
******************************************************************************/

void
print_tree (tree t, int tab=0) {
  int i;
  for (i=0; i<tab; i++) cout << " ";
  if (is_atomic (t)) cout << t->label << "\n";
  else {
    cout << CONSTRUCTOR_NAME [L(t)] << "\n";
    for (i=0; i<N(t); i++) print_tree (t[i], tab+2);
  }
}

tree
xml_html_parser::parse (string s2) {
  // end of line handling
  s= ""; i= 0; n= N(s2);
  bool is_cr= false;
  while (i<n) {
    bool prev_is_cr= is_cr;
    is_cr= false;
    char c= s2[i];
    if (c == '\15') {
      s << '\12';
      is_cr= true;
    }
    else if (prev_is_cr && (c == '\12')) /* no-op */;
    else s << c;
    i++;
  }

  // real parsing
  i= 0; n= N(s);
  // cout << "Parsing " << s << "\n";
  parse ();
  // cout << HRULE << LF;
  // cout << "a= " << a << "\n";
  i= 0; n= N(a); stack= tuple ("<bottom>");
  tree r= tuple ("tag", "<document>");
  build (r);
  // cout << HRULE << LF;
  // print_tree (r);
  r= finalize_entities (r);
  // cout << HRULE << LF;
  // print_tree (r);
  r= finalize_sxml (r);
  // cout << HRULE << LF;
  // print_tree (r);
  return r;
}

/******************************************************************************
* Interface
******************************************************************************/

tree
parse_html (string s) {
  xml_html_parser parser;
  parser.html= true;
  tree t= parser.parse (s);
  return t;
}
