
/******************************************************************************
* MODULE     : image_files.cc
* DESCRIPTION: image file handling
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "file.hh"
#include "image_files.hh"
#include "web_files.hh"
#include "sys_utils.hh"
#include "analyze.hh"

hashmap<tree,string> ps_bbox ("");

/******************************************************************************
* Loading xpm pixmaps
******************************************************************************/

tree
xpm_load (url u) {
  string s;
  load_string ("$TEXMACS_PIXMAPS_PATH" * u, s);
  if (s == "") load_string ("$TEXMACS_PATH/misc/pixmaps/TeXmacs.xpm", s);

  int i, j;
  tree t (TUPLE);
  for (i=0; i<N(s); i++)
    if (s[i]=='\x22') {
      i++;
      j=i;
      while ((i<N(s)) && (s[i]!='\x22')) i++;
      t << s (j, i);
    }
  if (N(t)==0) return xpm_load ("$TEXMACS_PATH/misc/pixmaps/TeXmacs.xpm");
  return t;
}

void
xpm_size (url u, int& w, int& h) {
  static hashmap<string,string> xpm_size_table ("");
  string file_name= as_string (u);
  if (!xpm_size_table->contains (file_name)) {
    tree t= xpm_load (u);
    xpm_size_table (file_name)= t[0]->label;
  }

  int i= 0;
  bool ok;
  string s= xpm_size_table[file_name];
  skip_spaces (s, i);
  ok= read_int (s, i, w);
  skip_spaces (s, i);
  ok= read_int (s, i, h) && ok;
  if (!ok) fatal_error ("Invalid xpm (" * file_name * ")", "xpm_size");
}

/******************************************************************************
* Loading postscript files or conversion to postscript
******************************************************************************/

static string
load_postscript (url file_name) {
  string s;
  load_string (file_name, s);
  if (s == "") load_string ("$TEXMACS_PATH/misc/pixmaps/unknown.ps", s);
  return s;
}

static string
load_tiff (url file_name) {
  return eval_system ("tiff2ps", file_name);
}

static string
load_pdf (url file_name) {
  url temp_name= url_temp ();
  system ("pdf2ps", file_name, temp_name);
  string s= load_postscript (temp_name);
  remove (temp_name);
  return s;
}

static string
load_pnm (url file_name) {
  return eval_system ("pnmtops -noturn", file_name);
}

static string
load_png (url file_name) {
  url temp_name= url_temp ();
  system ("pngtopnm", file_name, ">", temp_name);
  string s= load_pnm (temp_name);
  remove (temp_name);
  return s;
}

static string
load_gif (url file_name) {
  url temp_name= url_temp ();
  system ("giftopnm", file_name, "| cat >", temp_name);
  string s= load_pnm (temp_name);
  remove (temp_name);
  return s;
}

static string
load_ppm (url file_name) {
  url temp_name= url_temp ();
  system ("ppmtogif", file_name, ">", temp_name);
  string s= load_gif (temp_name);
  remove (temp_name);
  return s;
}

static string
load_jpg (url file_name) {
  url temp_name= url_temp ();
  system ("djpeg -pnm", file_name, ">", temp_name);
  string s= load_pnm (temp_name);
  remove (temp_name);
  return s;
}

static string
load_xpm (url file_name) {
  url temp_name= url_temp ();
  system ("xpmtoppm", file_name, ">", temp_name);
  string s= load_ppm (temp_name);
  remove (temp_name);
  return s;
}

static string
load_fig (url file_name) {
  url temp_name= url_temp ();
  system ("fig2ps", file_name, temp_name);
  string s= load_postscript (temp_name);
  remove (temp_name);
  return s;
}

string
ps_load (url image) {
  url file_name= resolve (image);
  if (is_none (file_name))
    file_name= "$TEXMACS_PATH/misc/pixmaps/unknown.ps";
  string s= suffix (file_name);
  if (s == "ps") return load_postscript (file_name);
  else if (s == "eps") return load_postscript (file_name);
  else if (s == "tif") return load_tiff (file_name);
  else if (s == "pdf") return load_pdf (file_name);
  else if (s == "pnm") return load_pnm (file_name);
  else if (s == "png") return load_png (file_name);
  else if (s == "gif") return load_gif (file_name);
  else if (s == "ppm") return load_ppm (file_name);
  else if (s == "jpg") return load_jpg (file_name);
  else if (s == "xpm") return load_xpm (file_name);
  else if (s == "fig") return load_fig (file_name);
  else return "";
}

void
ps_bounding_box (url image, int& x1, int& y1, int& x2, int& y2) {
  tree lookup= image->t;
  if (!ps_bbox->contains (image->t)) {
    int i;
    string s= ps_load (image);
    string r= "0 0 32 32";
    for (i=0; i<N(s); i++)
      if (read (s, i, "%%BoundingBox: ")) {
	read_line (s, i, r);
	break;
      }
    ps_bbox (image->t)= r;
  }

  int i= 0;
  bool ok;
  string s= ps_bbox [image->t];
  double X1=0.0, Y1=0.0, X2=0.0, Y2=0.0;
  skip_spaces (s, i);
  ok= read_double (s, i, X1);
  skip_spaces (s, i);
  ok= read_double (s, i, Y1) && ok;
  skip_spaces (s, i);
  ok= read_double (s, i, X2) && ok;
  skip_spaces (s, i);
  ok= read_double (s, i, Y2) && ok;
  x1= (int) X1; y1= (int) Y1;
  x2= (int) X2; y2= (int) Y2;
  if (ok) return;
  x1= y1= 0; x2= 596; y2= 842;
}
